"""
The definitions used within the package
"""

import re
from typing import Union
from dataclasses import dataclass

from .constants import ENTITY_TYPES

# ================================================
# Entity Definitions
# ================================================


@dataclass
class Entity:
    text: str
    label: str
    start_index: int
    end_index: int
    score: float = 1.0
    type: ENTITY_TYPES = None
    regex: Union[str, re.Pattern] = ".*"



from .interface import GeneratorInterface
from .llm_label_generator import LLMLabelGenerator
from .mask_label_generator import MaskLabelGenerator
from .number_generator import NumberGenerator
from .date_generator import DateGenerator

__all__ = [
    "GeneratorInterface",
    "LLMLabelGenerator",
    "MaskLabelGenerator",
    "NumberGenerator",
    "DateGenerator",
]



import re
import importlib
from typing import List, Tuple
import warnings

import torch
from spacy import displacy
from spacy.tokens import Doc
from gliner_spacy.pipeline import GlinerSpacy

from ..helpers import convert_spacy_to_entity
from ..regex import regex_map
from ...constants import LANGUAGES
from ...definitions import Entity

from .interface import ExtractorInterface


class EntityExtractor(ExtractorInterface):

    def __init__(
        self,
        labels: List[dict],
        lang: LANGUAGES = LANGUAGES.ENGLISH,
        score_th=0.5,
        use_gpu=False,
    ):
        self.lang = lang
        self.score_th = score_th
        self.use_gpu = use_gpu
        self.labels = self._prepare_labels(labels)
        self.pipeline = self._prepare_pipeline()

    def __call__(self, text: str) -> Tuple[Doc, List[Entity]]:
        doc = self.pipeline(text)
        entities, doc.ents = self._prepare_entities(doc)
        return doc, entities

    def display(self, doc: Doc):
        options = {"colors": {l["label"]: "#5C7AEA" for l in self.labels}}
        displacy.render(doc, style="ent", options=options)

    # ===========================================
    # Private methods
    # ===========================================

    def _prepare_labels(self, labels):
        for l in labels:
            if "regex" in l:
                continue
            regex = regex_map(l["type"])
            if regex is not None:
                l["regex"] = regex
        return labels

    def _create_gliner_config(self):
        map_location = "cpu"
        if self.use_gpu and not torch.cuda.is_available():
            return warnings.warn(
                "The user requested GPU use, but not available GPU was found. Reverting back to CPU use."
            )
        if self.use_gpu and torch.cuda.is_available():
            map_location = "cuda"

        return {
            # the model is specialized for extracting PII data
            "gliner_model": "urchade/gliner_multi_pii-v1",
            "labels": [l["label"] for l in self.labels],
            "threshold": self.score_th,
            "chunk_size": 384,
            "style": "ent",
            "map_location": map_location,
        }

    def _prepare_pipeline(self):
        # load the appropriate parser for the language
        module_lang, class_lang = self.lang[0].lower(), self.lang[1].lower().title()
        language_module = importlib.import_module(f"spacy.lang.{module_lang}")
        language_class = getattr(language_module, class_lang)
        # initialize the language parser
        nlp = language_class()
        nlp.add_pipe("sentencizer")
        gliner_config = self._create_gliner_config()
        nlp.add_pipe("gliner_spacy", config=gliner_config)
        return nlp

    def _prepare_entities(self, doc):
        # prepares the anonymized and spacy entities

        # TODO: make this part more generic
        anoni_entities = []
        spacy_entities = []
        for e in doc.ents:
            label = list(filter(lambda x: x["label"] == e.label_, self.labels))[0]
            if re.match(label["regex"], e.text):
                anoni_entities.append(convert_spacy_to_entity(e, **label))
                spacy_entities.append(e)
        return anoni_entities, spacy_entities



"""
Contains the pseudonymization strategy
"""

from typing import List

from .interface import StrategyInterface
from ...definitions import Entity


# =====================================
# Main class
# =====================================


class PseudonymizationStrategy(StrategyInterface):

    def __init__(self, mapping, *args, **kwargs):
        self.mapping = mapping

    def anonymize(self, text: str, entities: List[Entity], *args, **kwargs):
        replacements = []
        for ent in entities[::-1]:
            r = self._create_replacement(ent, text, replacements)
            text = (
                text[: r["start_index"]] + r["anonymized_text"] + text[r["end_index"] :]
            )
            replacements.append(r)
        return text, replacements[::-1]

    def _create_replacement(self, entity: Entity, text: str, replacements: List[dict]):
        # check if the replacement already exists
        anonymized_text = self._check_replacement(entity, replacements)
        # create a new replacement if it doesn't exist
        anonymized_text = (
            self.mapping(text, entity) if not anonymized_text else anonymized_text
        )
        return {
            "original_text": entity.text,
            "label": entity.label,
            "start_index": entity.start_index,
            "end_index": entity.end_index,
            "anonymized_text": anonymized_text,
        }

    def _check_replacement(self, entity: Entity, replacements: List[dict]):
        existing_replacement = list(
            filter(lambda x: x["original_text"] == entity.text, replacements)
        )
        return (
            existing_replacement[0]["anonymized_text"]
            if len(existing_replacement) > 0
            else None
        )



import unittest

from anonipy.utils.file_system import open_file

# =====================================
# Helper functions
# =====================================

from test.resources.example_outputs import WORD_TEXT, PDF_TEXT, TXT_TEXT

resources = {
    "word": "./test/resources/example.docx",
    "pdf": "./test/resources/example.pdf",
    "txt": "./test/resources/example.txt",
}

# =====================================
# Test Entity Extractor
# =====================================


class TestFileSystem(unittest.TestCase):
    def test_open_file_word(self):
        self.assertEqual(open_file(resources["word"]), WORD_TEXT)

    def test_open_file_pdf(self):
        self.assertEqual(open_file(resources["pdf"]), PDF_TEXT)

    def test_open_file_txt(self):
        self.assertEqual(open_file(resources["txt"]), TXT_TEXT)


if __name__ == "__main__":
    unittest.main()



import unittest

from anonipy.anonymize.regex import (
    regex_map,
    RegexMap,
    REGEX_STRING,
    REGEX_INTEGER,
    REGEX_FLOAT,
    REGEX_DATE,
    REGEX_EMAIL_ADDRESS,
    REGEX_PHONE_NUMBER,
    REGEX_WEBSITE_URL,
)
from anonipy.constants import ENTITY_TYPES

# =====================================
# Test Entity
# =====================================


class TestRegex(unittest.TestCase):

    def test_init(self):
        self.assertEqual(regex_map.__class__, RegexMap)
        self.assertEqual(hasattr(regex_map, "regex_mapping"), True)

    def test_regex_mapping(self):

        test_cases = [
            {"value": "string", "entity": ENTITY_TYPES.STRING, "regex": REGEX_STRING},
            {
                "value": "integer",
                "entity": ENTITY_TYPES.INTEGER,
                "regex": REGEX_INTEGER,
            },
            {"value": "float", "entity": ENTITY_TYPES.FLOAT, "regex": REGEX_FLOAT},
            {"value": "date", "entity": ENTITY_TYPES.DATE, "regex": REGEX_DATE},
            {
                "value": "email",
                "entity": ENTITY_TYPES.EMAIL,
                "regex": REGEX_EMAIL_ADDRESS,
            },
            {
                "value": "phone_number",
                "entity": ENTITY_TYPES.PHONE_NUMBER,
                "regex": REGEX_PHONE_NUMBER,
            },
            {
                "value": "website_url",
                "entity": ENTITY_TYPES.WEBSITE_URL,
                "regex": REGEX_WEBSITE_URL,
            },
            {"value": "custom", "entity": ENTITY_TYPES.CUSTOM, "regex": ".*"},
            {"value": "test", "entity": "test", "regex": ".*"},
        ]

        for test_case in test_cases:
            self.assertEqual(regex_map(test_case["entity"]), test_case["regex"])
            self.assertEqual(
                regex_map(test_case["value"]), regex_map(test_case["entity"])
            )


if __name__ == "__main__":
    unittest.main()



import re
import unittest
import warnings

import torch

from anonipy.definitions import Entity
from anonipy.anonymize.generators import (
    LLMLabelGenerator,
    MaskLabelGenerator,
    DateGenerator,
    NumberGenerator,
)

# =====================================
# Helper functions
# =====================================


original_text = """\
Medical Record

Patient Name: John Doe
Date of Birth: 15-01-1985
Date of Examination: 20-05-2024
Social Security Number: 123-45-6789

Examination Procedure:
John Doe underwent a routine physical examination. The procedure included measuring vital signs (blood pressure, heart rate, temperature), a comprehensive blood panel, and a cardiovascular stress test. The patient also reported occasional headaches and dizziness, prompting a neurological assessment and an MRI scan to rule out any underlying issues.

Medication Prescribed:

Ibuprofen 200 mg: Take one tablet every 6-8 hours as needed for headache and pain relief.
Lisinopril 10 mg: Take one tablet daily to manage high blood pressure.
Next Examination Date:
15-11-2024
"""

test_entities = {
    "name": Entity(
        text="John Doe",
        label="name",
        start_index=30,
        end_index=38,
        score=1.0,
        type="string",
        regex=".*",
    ),
    "date": Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        score=1.0,
        type="date",
        regex="(\\d{1,2}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{2,4})|(\\d{2,4}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{1,2})",
    ),
    "integer": Entity(
        text="123456789",
        label="integer",
        start_index=121,
        end_index=132,
        score=1.0,
        type="integer",
        regex="\d+",
    ),
    "float": Entity(
        text="123,456,789.000",
        label="float",
        start_index=121,
        end_index=132,
        score=1.0,
        type="float",
        regex="[\d\.,]+",
    ),
    "custom": Entity(
        text="123-45-6789",
        label="custom",
        start_index=121,
        end_index=132,
        score=1.0,
        type="custom",
        regex="\\d{3}-\\d{2}-\\d{4}",
    ),
}


# =====================================
# Test LLM Label Generator
# =====================================

if torch.cuda.is_available():
    # ! THESE TESTS REQUIRE GPU/CUDA! THIS WILL FAIL IF YOU DON'T HAVE GPU/CUDA!

    class TestLLMLabelGenerator(unittest.TestCase):

        @classmethod
        def setUpClass(self):
            self.generator = LLMLabelGenerator()

        def test_has_methods(self):
            self.assertEqual(hasattr(self.generator, "generate"), True)

        def test_generate_default(self):
            entity = test_entities["name"]
            generated_text = self.generator.generate(entity)
            match = re.match(entity.regex, generated_text)
            self.assertNotEqual(match, None)
            self.assertEqual(match.group(0), generated_text)

        def test_generate_custom(self):
            entity = test_entities["name"]
            generated_text = self.generator.generate(
                entity, entity_prefix="Spanish", temperature=0.5
            )
            match = re.match(entity.regex, generated_text)
            self.assertNotEqual(match, None)
            self.assertEqual(match.group(0), generated_text)


# =====================================
# Test Mask Label Generator
# =====================================


class TestMaskLabelGenerator(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.generator = MaskLabelGenerator()

    def setUp(self):
        warnings.filterwarnings("ignore", category=ImportWarning)
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.filterwarnings("ignore", category=FutureWarning)

    def test_has_methods(self):
        self.assertEqual(hasattr(self.generator, "generate"), True)

    def test_generate_default(self):
        entity = test_entities["name"]
        generated_text = self.generator.generate(entity, text=original_text)
        match = re.match(entity.regex, generated_text)
        self.assertNotEqual(match, None)
        self.assertEqual(match.group(0), generated_text)


# =====================================
# Test Date Generator
# =====================================


class TestDateGenerator(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.generator = DateGenerator()

    def test_has_methods(self):
        self.assertEqual(hasattr(self.generator, "generate"), True)

    def test_generate_default(self):
        entity = test_entities["date"]
        generated_text = self.generator.generate(entity)
        match = re.match(entity.regex, generated_text)
        self.assertNotEqual(match, None)
        self.assertEqual(match.group(0), generated_text)

    def test_generate_first_day_of_the_month(self):
        entity = test_entities["date"]
        generated_text = self.generator.generate(
            entity, output_gen="first_day_of_the_month"
        )
        self.assertEqual(generated_text, "01-05-2024")

    def test_generate_last_day_of_the_month(self):
        entity = test_entities["date"]
        generated_text = self.generator.generate(
            entity, output_gen="last_day_of_the_month"
        )
        self.assertEqual(generated_text, "31-05-2024")

    def test_generate_middle_of_the_month(self):
        entity = test_entities["date"]
        generated_text = self.generator.generate(
            entity, output_gen="middle_of_the_month"
        )
        self.assertEqual(generated_text, "15-05-2024")

    def test_generate_middle_of_the_year(self):
        entity = test_entities["date"]
        generated_text = self.generator.generate(
            entity, output_gen="middle_of_the_year"
        )
        self.assertEqual(generated_text, "01-07-2024")

    def test_generate_random(self):
        entity = test_entities["date"]
        generated_text = self.generator.generate(entity, output_gen="random")
        match = re.match(entity.regex, generated_text)
        self.assertNotEqual(match, None)
        self.assertEqual(match.group(0), generated_text)

    def test_generate_uncorrect_type(self):
        entity = test_entities["name"]
        try:
            self.generator.generate(entity)
        except Exception as e:
            self.assertEqual(type(e), ValueError)


# =====================================
# Test Number Generator
# =====================================


class TestNumberGenerator(unittest.TestCase):

    @classmethod
    def setUpClass(self):
        self.generator = NumberGenerator()

    def test_has_methods(self):
        self.assertEqual(hasattr(self.generator, "generate"), True)

    def test_generate_integer(self):
        entity = test_entities["integer"]
        generated_text = self.generator.generate(entity)
        match = re.match(entity.regex, generated_text)
        self.assertNotEqual(match, None)
        self.assertEqual(match.group(0), generated_text)

    def test_generate_float(self):
        entity = test_entities["float"]
        generated_text = self.generator.generate(entity)
        match = re.match(entity.regex, generated_text)
        self.assertNotEqual(match, None)
        self.assertEqual(match.group(0), generated_text)

    def test_generate_custom(self):
        entity = test_entities["custom"]
        generated_text = self.generator.generate(entity)
        match = re.match(entity.regex, generated_text)
        self.assertNotEqual(match, None)
        self.assertEqual(match.group(0), generated_text)

    def test_generate_uncorrect_type(self):
        entity = test_entities["name"]
        try:
            self.generator.generate(entity)
        except Exception as e:
            self.assertEqual(type(e), ValueError)


if __name__ == "__main__":
    unittest.main()



import unittest

from anonipy.definitions import Entity


# =====================================
# Test Entity
# =====================================


class TestEntity(unittest.TestCase):

    def test_init_default(self):
        entity = Entity(
            text="test",
            label="test",
            start_index=0,
            end_index=4,
        )
        self.assertEqual(entity.text, "test")
        self.assertEqual(entity.label, "test")
        self.assertEqual(entity.start_index, 0)
        self.assertEqual(entity.end_index, 4)
        self.assertEqual(entity.score, 1.0)
        self.assertEqual(entity.type, None)
        self.assertEqual(entity.regex, ".*")

    def test_init_custom(self):
        entity = Entity(
            text="test",
            label="test",
            start_index=0,
            end_index=4,
            score=0.89,
            type="test",
            regex="test",
        )
        self.assertEqual(entity.text, "test")
        self.assertEqual(entity.label, "test")
        self.assertEqual(entity.start_index, 0)
        self.assertEqual(entity.end_index, 4)
        self.assertEqual(entity.score, 0.89)
        self.assertEqual(entity.type, "test")
        self.assertEqual(entity.regex, "test")


if __name__ == "__main__":
    unittest.main()



import random
import warnings
import datetime
from .interface import GeneratorInterface
from ...definitions import Entity

# =====================================
# Helper functions
# =====================================


def first_day_of_month(day: datetime.datetime, *args, **kwargs):
    return day.replace(day=1)


def last_day_of_month(day: datetime.datetime, *args, **kwargs):
    next_month = day.replace(day=28) + datetime.timedelta(days=4)
    return next_month - datetime.timedelta(days=next_month.day)


def middle_of_the_month(day: datetime.datetime, *args, **kwargs):
    return day.replace(day=15)


def middle_of_the_year(day: datetime.datetime, *args, **kwargs):
    return day.replace(month=7, day=1)


def random_date(day: datetime.datetime, sigma: int = 30, *args, **kwargs):
    delta = random.randint(-sigma, sigma)
    return day + datetime.timedelta(days=delta)


operations = {
    "first_day_of_the_month": first_day_of_month,
    "last_day_of_the_month": last_day_of_month,
    "middle_of_the_month": middle_of_the_month,
    "middle_of_the_year": middle_of_the_year,
    "random": random_date,
}


# =====================================
# Main class
# =====================================


class DateGenerator(GeneratorInterface):

    def __init__(
        self, date_format: str = "%d-%m-%Y", day_sigma: int = 30, *args, **kwargs
    ):
        self.date_format = date_format
        self.day_sigma = day_sigma

    def generate(self, entity: Entity, output_gen: str = "random", *args, **kwargs):
        if entity.type in ["custom"]:
            warnings.warn(
                "The entity type is `custom`. Make sure the generator is returning appropriate values."
            )
        elif entity.type not in ["date"]:
            raise ValueError("The entity type must be `date` to generate dates.")

        if output_gen not in operations.keys():
            raise ValueError(
                f"The output_gen must be one of {', '.join(list(operations.keys()))} to generate dates."
            )

        entity_date = datetime.datetime.strptime(entity.text, self.date_format)
        generate_date = operations[output_gen](entity_date, self.day_sigma)
        return generate_date.strftime(self.date_format)



"""
Contains the redaction strategy
"""

from typing import List

from .interface import StrategyInterface
from ...definitions import Entity


# =====================================
# Main class
# =====================================


class RedactionStrategy(StrategyInterface):

    def __init__(self, substitute_label: str = "[REDACTED]", *args, **kwargs):
        self.substitute_label = substitute_label

    def anonymize(self, text: str, entities: List[Entity], *args, **kwargs):
        replacements = []
        for ent in entities[::-1]:
            r = self._create_replacement(ent)
            text = (
                text[: r["start_index"]] + r["anonymized_text"] + text[r["end_index"] :]
            )
            replacements.append(r)
        return text, replacements[::-1]

    def _create_replacement(self, entity: Entity):
        return {
            "original_text": entity.text,
            "label": entity.label,
            "start_index": entity.start_index,
            "end_index": entity.end_index,
            "anonymized_text": self.substitute_label,
        }



from ...definitions import Entity


# =====================================
# Main functions
# =====================================


class GeneratorInterface:

    def __init__(self, *args, **kwargs):
        pass

    def generate(self, entity: Entity, *args, **kwargs):
        pass



import torch
from tokenizers import pre_tokenizers
from transformers import AutoModelForCausalLM, AutoTokenizer, BitsAndBytesConfig

from guidance import models, gen, select

from .interface import GeneratorInterface
from ...definitions import Entity


# =====================================
# Helper functions
# =====================================


def prepare_llama3_byte_decoder():
    """Prepares the byte decoder

    This is an implementation of a workaround, such that the guidance module
    can be used with the LLaMa-3 model from Hugging Face. Once the issue is resolved,
    we will remove this function.

    Link to the guidance issue: https://github.com/guidance-ai/guidance/issues/782
    """
    byte_decoder = {}
    # alphabet = pre_tokenizers.ByteLevel(False, False).alphabet()
    known_vals = set([])

    for j in range(256):
        for k in range(256):
            for l in range(256):
                if len(byte_decoder.keys()) < 256:
                    b = b""
                    vals = [j, k, l]
                    if not set(vals).issubset(known_vals):
                        for d in range(3):
                            b = b + vals[d].to_bytes(1, "little", signed=False)
                        try:
                            c = b.decode()
                            t = pre_tokenizers.ByteLevel(False, False).pre_tokenize_str(
                                c
                            )[0][0]
                            for m in range(3):
                                if t[m] not in byte_decoder.keys():
                                    byte_decoder[t[m]] = vals[m]
                                    known_vals.add(vals[m])
                        except UnicodeDecodeError:
                            pass

    byte_decoder["À"] = 192
    byte_decoder["Á"] = 193
    byte_decoder["ð"] = 240
    byte_decoder["ñ"] = 241
    byte_decoder["ò"] = 242
    byte_decoder["ó"] = 243
    byte_decoder["ô"] = 244
    byte_decoder["õ"] = 245
    byte_decoder["ö"] = 246
    byte_decoder["÷"] = 247
    byte_decoder["ø"] = 248
    byte_decoder["ù"] = 249
    byte_decoder["ú"] = 250
    byte_decoder["û"] = 251
    byte_decoder["ü"] = 252
    byte_decoder["ý"] = 253
    byte_decoder["þ"] = 254
    byte_decoder["ÿ"] = 255

    return byte_decoder


# =====================================
# Main class
# =====================================


class LLMLabelGenerator(GeneratorInterface):

    def __init__(self, *args, **kwargs):
        # TODO: make this configurable
        model_name = "meta-llama/Meta-Llama-3-8B-Instruct"

        if not torch.cuda.is_available():
            raise RuntimeError(
                "The LabelGenerator requires GPU/CUDA, but it is not available."
            )

        model, tokenizer = self._prepare_model_and_tokenizer(model_name)
        self.model = models.Transformers(model=model, tokenizer=tokenizer, echo=False)

    def generate(
        self,
        entity: Entity,
        entity_prefix: str = "",
        temperature: float = 0.0,
        *args,
        **kwargs,
    ):
        user_prompt = f"What is a random {entity_prefix} {entity.label} replacement for {entity.text}? Respond only with the replacement."
        assistant_prompt = gen(
            name="replacement",
            stop="<|eot_id|>",
            regex=entity.regex,
            temperature=temperature,
        )
        # generate the replacement for the entity
        lm = (
            self.model
            + self._system_prompt()
            + self._user_prompt(user_prompt)
            + self._assistant_prompt(assistant_prompt)
        )
        return lm["replacement"]

    def validate(self, entity: Entity):
        regex = regex if regex else ".*"
        user_prompt = f"Is {entity.text} a {entity.label}?"
        assistant_prompt = select(["True", "False"], name="validation")
        # validate the entity with the validation prompt
        lm = (
            self.model
            + self._system_prompt()
            + self._user_prompt(user_prompt)
            + self._assistant_prompt(assistant_prompt)
        )
        return bool(lm["validation"])

    # =================================
    # Private methods
    # =================================

    def _prepare_model_and_tokenizer(self, model_name: str):
        # prepare the model
        bnb_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_use_double_quant=True,
            bnb_4bit_compute_dtype=torch.bfloat16,
        )
        model = AutoModelForCausalLM.from_pretrained(
            model_name, quantization_config=bnb_config, low_cpu_mem_usage=True
        )
        # prepare the tokenizer
        tokenizer = AutoTokenizer.from_pretrained(
            model_name, padding_side="right", use_fast=False
        )
        # use the workaround for the LLaMa-3 model
        tokenizer.byte_decoder = prepare_llama3_byte_decoder()
        return model, tokenizer

    def _system_prompt(self):
        return "<|begin_of_text|><|start_header_id|>system<|end_header_id|>\n\nYou are a helpful AI assistant for generating replacements for text entities.<|eot_id|>"

    def _user_prompt(self, prompt):
        return f"<|start_header_id|>user<|end_header_id|>\n\n{prompt}<|eot_id|>"

    def _assistant_prompt(self, prompt):
        return f"<|start_header_id|>assistant<|end_header_id|>\n\n{prompt}"



"""
The language detector class
"""

from lingua import LanguageDetectorBuilder


class LanguageDetector:

    def __init__(self, low_accuracy: bool = False):
        # Prepare the language detector for all languages
        builder = LanguageDetectorBuilder.from_all_languages()
        builder = (
            builder.with_low_accuracy_mode()
            if low_accuracy
            else builder.with_preloaded_language_models()
        )
        self.detector = builder.build()

    def __call__(self, text: str, output_standard: str = "iso_code_639_1") -> str:
        return self.detect(text, output_standard)

    def detect(self, text: str, output_standard: str = "iso_code_639_1") -> str:
        language = self.detector.detect_language_of(text)
        iso_code = getattr(language, output_standard).name.lower()
        full_name = language.name.lower().title()
        return iso_code, full_name



"""
The regex definitions for various use cases
"""

from collections import defaultdict

from ..constants import ENTITY_TYPES


# =====================================
# Regex definitions
# =====================================

REGEX_STRING = ".*"
REGEX_INTEGER = "\d+"
REGEX_FLOAT = "[\d\.,]+"
REGEX_DATE = (
    "(\d{1,2}[\/\-\.]\d{1,2}[\/\-\.]\d{2,4})|(\d{2,4}[\/\-\.]\d{1,2}[\/\-\.]\d{1,2})"
)
REGEX_EMAIL_ADDRESS = (
    "[a-zA-Z0-9.!#$%&’*+/=?^_`{|}~-]+@[a-zA-Z0-9-]+(?:\.[a-zA-Z0-9-]+)*"
)
REGEX_PHONE_NUMBER = (
    "[(]?[\+]?[(]?[0-9]{1,3}[)]?[-\s\.]?([0-9]{2,}[-\s\.]?){2,}([0-9]{3,})"
)
REGEX_WEBSITE_URL = "((https?|ftp|smtp):\/\/)?(www.)?([a-zA-Z0-9]+\.)+[a-z]{2,}(\/[a-zA-Z0-9#\?\_\.\=\-\&]+|\/?)*"


# =====================================
# Define the regex definitions
# =====================================


class RegexMap:

    def __init__(self):
        self.regex_mapping = defaultdict(lambda: ".*")
        # Define the regex mappings
        self.regex_mapping[ENTITY_TYPES.STRING] = REGEX_STRING
        self.regex_mapping[ENTITY_TYPES.INTEGER] = REGEX_INTEGER
        self.regex_mapping[ENTITY_TYPES.FLOAT] = REGEX_FLOAT
        self.regex_mapping[ENTITY_TYPES.DATE] = REGEX_DATE
        self.regex_mapping[ENTITY_TYPES.EMAIL] = REGEX_EMAIL_ADDRESS
        self.regex_mapping[ENTITY_TYPES.PHONE_NUMBER] = REGEX_PHONE_NUMBER
        self.regex_mapping[ENTITY_TYPES.WEBSITE_URL] = REGEX_WEBSITE_URL

    def __call__(self, type: str) -> str:
        return self.regex_mapping[type]


regex_map = RegexMap()



from ..definitions import Entity

# =====================================
# Entity converters
# =====================================


def convert_spacy_to_entity(entity, type=None, regex=".*", *args, **kwargs):
    return Entity(
        entity.text,
        entity.label_,
        entity.start_char,
        entity.end_char,
        entity._.score,
        type,
        regex,
    )



"""
Contains the interface for the strategy
"""

from typing import List
from ...definitions import Entity

# =====================================
# Main class
# =====================================


class StrategyInterface:

    def __init__(self, *args, **kwargs):
        pass

    def anonymize(self, text: str, entities: List[Entity], *args, **kwargs):
        pass



"""
Contains the masking strategy
"""

import re
from typing import List

from .interface import StrategyInterface
from ...definitions import Entity


# =====================================
# Main class
# =====================================


class MaskingStrategy(StrategyInterface):

    def __init__(self, substitute_label: str = "*", *args, **kwargs):
        self.substitute_label = substitute_label

    def anonymize(self, text: str, entities: List[Entity], *args, **kwargs):
        replacements = []
        for ent in entities[::-1]:
            r = self._create_replacement(ent)
            text = (
                text[: r["start_index"]] + r["anonymized_text"] + text[r["end_index"] :]
            )
            replacements.append(r)
        return text, replacements[::-1]

    def _create_replacement(self, entity: Entity):
        mask = self._create_mask(entity)
        return {
            "original_text": entity.text,
            "label": entity.label,
            "start_index": entity.start_index,
            "end_index": entity.end_index,
            "anonymized_text": mask,
        }

    def _create_mask(self, entity: Entity):
        return " ".join(
            [
                self.substitute_label * len(chunk)
                for chunk in re.split(r"\s+", entity.text)
            ]
        )



"""
The constants used to make it easier to use the library
"""

# ================================================
# Constants
# ================================================


class LANGUAGES:
    """Supported languages"""

    DUTCH = ("nl", "Dutch")
    ENGLISH = ("en", "English")
    FRENCH = ("fr", "French")
    GERMAN = ("de", "German")
    GREEK = ("el", "Greek")
    ITALIAN = ("it", "Italian")
    SLOVENE = ("sl", "Slovene")
    SPANISH = ("es", "Spanish")
    UKRAINIAN = ("uk", "Ukrainian")


class ENTITY_TYPES:
    """Types of entities"""

    CUSTOM = "custom"
    STRING = "string"
    INTEGER = "integer"
    FLOAT = "float"
    DATE = "date"
    EMAIL = "email"
    WEBSITE_URL = "website_url"
    PHONE_NUMBER = "phone_number"



import random
import warnings
from .interface import GeneratorInterface
from ...definitions import Entity

# =====================================
# Main class
# =====================================


class NumberGenerator(GeneratorInterface):

    def __init__(self, *args, **kwargs):
        pass

    def generate(self, entity: Entity, *args, **kwargs):
        if entity.type in ["custom"]:
            warnings.warn(
                "The entity type is `custom`. Make sure the generator is returning appropriate values."
            )
        elif entity.type not in ["integer", "float", "phone_number"]:
            raise ValueError(
                "The entity type must be `integer`, `float`, `phone_number` or `custom` to generate numbers."
            )
        return "".join(
            [str(random.randint(0, 9)) if d.isdigit() else d for d in entity.text]
        )



"""
anonipy
=========

Provides
  1. Label extractors
  2. Label generators
  3. Anonymization strategies

How to use the documentation
----------------------------
Documentation is available in two forms: docstrings provided
with the code and a loose standing reference guide, available
from `the anonipy homepage <https://eriknovak.github.io/anonipy>`.

"""

__version__ = "0.0.6"

from . import anonymize
from . import utils
from . import definitions
from . import constants



import unittest

from anonipy.definitions import Entity
from anonipy.anonymize.strategies import (
    RedactionStrategy,
    MaskingStrategy,
    PseudonymizationStrategy,
)

# =====================================
# Helper functions
# =====================================

test_text = "Test this string, and this test too!"
test_entities = [
    Entity(text="Test", label="test", start_index=0, end_index=4, score=1.0),
    Entity(text="string", label="type", start_index=10, end_index=16, score=1.0),
    Entity(text="test", label="test", start_index=27, end_index=31, score=1.0),
]


def anonymization_mapping(text, entity):
    if entity.label == "test":
        return "[TEST]"
    elif entity.label == "type":
        return "[TYPE]"
    return "[REDACTED]"


# =====================================
# Test Redaction Strategy
# =====================================


class TestRedactionStrategy(unittest.TestCase):
    def test_init(self):
        strategy = RedactionStrategy()
        self.assertEqual(strategy.__class__, RedactionStrategy)

    def test_has_methods(self):
        strategy = RedactionStrategy()
        self.assertEqual(hasattr(strategy, "anonymize"), True)

    def test_default_inputs(self):
        strategy = RedactionStrategy()
        self.assertEqual(strategy.substitute_label, "[REDACTED]")

    def test_custom_inputs(self):
        strategy = RedactionStrategy(substitute_label="[TEST]")
        self.assertEqual(strategy.substitute_label, "[TEST]")

    def test_anonymize_default_inputs(self):
        strategy = RedactionStrategy()
        anonymized_text, replacements = strategy.anonymize(test_text, test_entities)
        self.assertEqual(
            anonymized_text, "[REDACTED] this [REDACTED], and this [REDACTED] too!"
        )
        self.assertEqual(
            replacements,
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "[REDACTED]",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "[REDACTED]",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "[REDACTED]",
                },
            ],
        )

    def test_anonymize_custom_inputs(self):
        strategy = RedactionStrategy(substitute_label="[TEST]")
        anonymized_text, replacements = strategy.anonymize(test_text, test_entities)
        self.assertEqual(anonymized_text, "[TEST] this [TEST], and this [TEST] too!")
        self.assertEqual(
            replacements,
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "[TEST]",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "[TEST]",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "[TEST]",
                },
            ],
        )


# =====================================
# Test Masking Strategy
# =====================================


class TestMaskingStrategy(unittest.TestCase):
    def test_init(self):
        strategy = MaskingStrategy()
        self.assertEqual(strategy.__class__, MaskingStrategy)

    def test_methods(self):
        strategy = MaskingStrategy()
        self.assertEqual(hasattr(strategy, "anonymize"), True)

    def test_default_inputs(self):
        strategy = MaskingStrategy()
        self.assertEqual(strategy.substitute_label, "*")

    def test_custom_inputs(self):
        strategy = MaskingStrategy(substitute_label="A")
        self.assertEqual(strategy.substitute_label, "A")

    def test_anonymize_default_inputs(self):
        strategy = MaskingStrategy()
        anonymized_text, replacements = strategy.anonymize(test_text, test_entities)
        self.assertEqual(anonymized_text, "**** this ******, and this **** too!")
        self.assertEqual(
            replacements,
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "****",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "******",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "****",
                },
            ],
        )

    def test_anonymize_custom_inputs(self):
        strategy = MaskingStrategy(substitute_label="A")
        anonymized_text, replacements = strategy.anonymize(test_text, test_entities)
        self.assertEqual(anonymized_text, "AAAA this AAAAAA, and this AAAA too!")
        self.assertEqual(
            replacements,
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "AAAA",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "AAAAAA",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "AAAA",
                },
            ],
        )


# =====================================
# Test Pseudonymization Strategy
# =====================================


class TestPseudonymizationStrategy(unittest.TestCase):
    def test_init(self):
        try:
            PseudonymizationStrategy()
        except Exception as e:
            self.assertRaises(TypeError, e)

    def test_init_inputs(self):
        strategy = PseudonymizationStrategy(mapping=anonymization_mapping)
        self.assertEqual(strategy.__class__, PseudonymizationStrategy)

    def test_methods(self):
        strategy = PseudonymizationStrategy(mapping=anonymization_mapping)
        self.assertEqual(hasattr(strategy, "anonymize"), True)

    def test_anonymize_inputs(self):
        strategy = PseudonymizationStrategy(mapping=anonymization_mapping)
        anonymized_text, replacements = strategy.anonymize(test_text, test_entities)
        self.assertEqual(anonymized_text, "[TEST] this [TYPE], and this [TEST] too!")
        self.assertEqual(
            replacements,
            [
                {
                    "original_text": "Test",
                    "label": "test",
                    "start_index": 0,
                    "end_index": 4,
                    "anonymized_text": "[TEST]",
                },
                {
                    "original_text": "string",
                    "label": "type",
                    "start_index": 10,
                    "end_index": 16,
                    "anonymized_text": "[TYPE]",
                },
                {
                    "original_text": "test",
                    "label": "test",
                    "start_index": 27,
                    "end_index": 31,
                    "anonymized_text": "[TEST]",
                },
            ],
        )


if __name__ == "__main__":
    unittest.main()



import unittest
import warnings

import torch

from anonipy.definitions import Entity
from anonipy.anonymize.extractors import EntityExtractor
from anonipy.constants import LANGUAGES

# =====================================
# Helper functions
# =====================================

original_text = """\
Medical Record

Patient Name: John Doe
Date of Birth: 15-01-1985
Date of Examination: 20-05-2024
Social Security Number: 123-45-6789

Examination Procedure:
John Doe underwent a routine physical examination. The procedure included measuring vital signs (blood pressure, heart rate, temperature), a comprehensive blood panel, and a cardiovascular stress test. The patient also reported occasional headaches and dizziness, prompting a neurological assessment and an MRI scan to rule out any underlying issues.

Medication Prescribed:

Ibuprofen 200 mg: Take one tablet every 6-8 hours as needed for headache and pain relief.
Lisinopril 10 mg: Take one tablet daily to manage high blood pressure.
Next Examination Date:
15-11-2024
"""

original_entities = [
    Entity(
        text="John Doe",
        label="name",
        start_index=30,
        end_index=38,
        type="string",
        regex=".*",
    ),
    Entity(
        text="15-01-1985",
        label="date of birth",
        start_index=54,
        end_index=64,
        type="date",
        regex="(\\d{1,2}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{2,4})|(\\d{2,4}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{1,2})",
    ),
    Entity(
        text="20-05-2024",
        label="date",
        start_index=86,
        end_index=96,
        type="date",
        regex="(\\d{1,2}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{2,4})|(\\d{2,4}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{1,2})",
    ),
    Entity(
        text="123-45-6789",
        label="social security number",
        start_index=121,
        end_index=132,
        type="custom",
        regex="[0-9]{3}-[0-9]{2}-[0-9]{4}",
    ),
    Entity(
        text="John Doe",
        label="name",
        start_index=157,
        end_index=165,
        type="string",
        regex=".*",
    ),
    Entity(
        text="15-11-2024",
        label="date",
        start_index=717,
        end_index=727,
        type="date",
        regex="(\\d{1,2}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{2,4})|(\\d{2,4}[\\/\\-\\.]\\d{1,2}[\\/\\-\\.]\\d{1,2})",
    ),
]

# define the labels to be extracted and anonymized
labels = [
    {"label": "name", "type": "string"},
    {
        "label": "social security number",
        "type": "custom",
        "regex": "[0-9]{3}-[0-9]{2}-[0-9]{4}",
    },
    {"label": "date of birth", "type": "date"},
    {"label": "date", "type": "date"},
]


# =====================================
# Test Entity Extractor
# =====================================


class TestEntityExtractor(unittest.TestCase):

    def setUp(self):
        warnings.filterwarnings("ignore", category=ImportWarning)
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.filterwarnings("ignore", category=FutureWarning)

    def test_init(self):
        try:
            EntityExtractor()
        except Exception as e:
            self.assertRaises(TypeError, e)

    def test_init_inputs(self):
        extractor = EntityExtractor(labels=labels, lang=LANGUAGES.ENGLISH, score_th=0.5)
        self.assertEqual(extractor.__class__, EntityExtractor)

    def test_init_gpu(self):
        if torch.cuda.is_available():
            extractor = EntityExtractor(
                labels=labels, lang=LANGUAGES.ENGLISH, score_th=0.5, use_gpu=True
            )
            self.assertEqual(extractor.__class__, EntityExtractor)

    def test_methods(self):
        extractor = EntityExtractor(labels=labels, lang=LANGUAGES.ENGLISH, score_th=0.5)
        self.assertEqual(hasattr(extractor, "__call__"), True)
        self.assertEqual(hasattr(extractor, "display"), True)

    def test_extract_default(self):
        extractor = EntityExtractor(labels=labels, lang=LANGUAGES.ENGLISH, score_th=0.5)
        doc, entities = extractor(original_text)
        for pred_entity, orig_entity in zip(entities, original_entities):
            self.assertEqual(pred_entity.text, orig_entity.text)
            self.assertEqual(pred_entity.label, orig_entity.label)
            self.assertEqual(pred_entity.start_index, orig_entity.start_index)
            self.assertEqual(pred_entity.end_index, orig_entity.end_index)
            self.assertEqual(pred_entity.type, orig_entity.type)
            self.assertEqual(pred_entity.regex, orig_entity.regex)
            self.assertEqual(pred_entity.score >= 0.5, True)


if __name__ == "__main__":
    unittest.main()



from .interface import StrategyInterface
from .masking import MaskingStrategy
from .redaction import RedactionStrategy
from .pseudonymization import PseudonymizationStrategy


__all__ = [
    "StrategyInterface",
    "MaskingStrategy",
    "RedactionStrategy",
    "PseudonymizationStrategy",
]



"""
The file system utilities
"""

import os
import re
import json
from typing import Union

from docx import Document
from pypdf import PdfReader


# Define namespaces
WORD_NAMESPACES = {"w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main"}

# =====================================
# Helper functions
# =====================================


def remove_extra_spaces(text: str) -> str:
    text = text.strip()
    # remove extra spaces
    text = re.sub(" +", " ", text)
    text = re.sub("\n{2,}", "\n\n", text)
    return text


def remove_page_numbers(text: str) -> str:
    page_number_pattern = re.compile(r"^\s*\d+\s*$|\s*\d+\s*$")
    filtered_lines = [
        line.strip()
        for line in text.splitlines()
        if not page_number_pattern.match(line)
    ]
    return "\n".join(filtered_lines)


# =====================================
# PDF extractor
# =====================================


def extract_text_from_pdf(pdf_path: str) -> str:
    pdf_reader = PdfReader(pdf_path)

    pages_text = []
    for page in pdf_reader.pages:
        text = page.extract_text(extraction_mode="layout")
        text = remove_page_numbers(text)
        text = remove_extra_spaces(text)
        pages_text.append(text)
    document_text = "\n".join(pages_text)

    return document_text


# =====================================
# Word extractor
# =====================================


def _word_process_paragraph(p) -> str:
    return p.text


def _word_process_table(t) -> str:
    table_text = []
    for row in t.findall(".//w:tr", WORD_NAMESPACES):
        row_text = []
        for cell in row.findall(".//w:tc", WORD_NAMESPACES):
            cell_text = []
            for p in cell.findall(".//w:p", WORD_NAMESPACES):
                cell_text.append(p.text)
            row_text.append(" ".join(cell_text))
        table_text.append(" ".join(row_text))
    return "\n".join(table_text)


def extract_text_from_word(doc_path: str) -> str:
    doc = Document(doc_path)
    content = []
    for element in doc.element.body:
        if element.tag.endswith("p"):
            # element is a paragraph
            text = _word_process_paragraph(element)
            content.append(text)
        elif element.tag.endswith("tbl"):
            # element is a table
            text = _word_process_table(element)
            content.append(text)
    document_text = "\n".join(content)
    return document_text


# =====================================
# Main functions
# =====================================


def open_file(file_path: str) -> str:
    _, file_extension = os.path.splitext(file_path)
    if file_extension.lower() == ".pdf":
        return extract_text_from_pdf(file_path)
    elif file_extension.lower() in [".doc", ".docx"]:
        return extract_text_from_word(file_path)
    elif file_extension.lower() == ".txt":
        with open(file_path, "r", encoding="utf-8") as f:
            return f.read()
    else:
        raise ValueError(f"The file extension is not supported: {file_extension}")


def open_json(file_path: str) -> dict:
    with open(file_path, "r", encoding="utf-8") as f:
        return json.load(f)


def write_file(text: str, file_path: str, encode: Union[str, bool] = True) -> None:
    if not isinstance(encode, str) and not isinstance(encode, bool):
        raise TypeError("encode must be a string or a boolean")

    encoding = None
    if isinstance(encode, str):
        encoding = encode
    elif isinstance(encode, bool):
        encoding = "utf-8" if encode else None

    with open(file_path, "w", encoding=encoding) as f:
        f.write(text)


def write_json(data: dict, file_path: str) -> None:
    with open(file_path, "w", encoding="utf-8") as f:
        json.dump(data, f, ensure_ascii=False, indent=4)



import unittest
import warnings

from anonipy.utils.language_detector import LanguageDetector
from anonipy.constants import LANGUAGES

# =====================================
# Test Language Detector
# =====================================


class TestLanguageDetector(unittest.TestCase):

    def setUp(self):
        warnings.filterwarnings("ignore", category=ImportWarning)
        warnings.filterwarnings("ignore", category=UserWarning)
        warnings.filterwarnings("ignore", category=FutureWarning)

    def test_init(self):
        language_detector = LanguageDetector()
        self.assertEqual(language_detector.__class__, LanguageDetector)

    def test_has_methods(self):
        language_detector = LanguageDetector()
        self.assertEqual(hasattr(language_detector, "detect"), True)

    def test_detect_english(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "This test verifies that the method is working correctly"
        )
        self.assertEqual(language[0], "en")
        self.assertEqual(language[1], "English")
        self.assertEqual(language, LANGUAGES.ENGLISH)

    def test_detect_slovene(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Ta test preverja, ali metoda dela pravilno"
        )
        self.assertEqual(language[0], "sl")
        self.assertEqual(language[1], "Slovene")
        self.assertEqual(language, LANGUAGES.SLOVENE)

    def test_detect_german(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Dieser Test überprüft, ob die Methode ordnungsgemäß funktioniert"
        )
        self.assertEqual(language[0], "de")
        self.assertEqual(language[1], "German")
        self.assertEqual(language, LANGUAGES.GERMAN)

    def test_detect_dutch(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Deze test verifieert dat de methode correct werkt"
        )
        self.assertEqual(language[0], "nl")
        self.assertEqual(language[1], "Dutch")
        self.assertEqual(language, LANGUAGES.DUTCH)

    def test_detect_spanish(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Esta prueba verifica que el método está funcionando correctamente"
        )
        self.assertEqual(language[0], "es")
        self.assertEqual(language[1], "Spanish")
        self.assertEqual(language, LANGUAGES.SPANISH)

    def test_detect_greek(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Αυτή η δοκιμή επαληθεύει ότι η μέθοδος λειτουργεί σωστά"
        )
        self.assertEqual(language[0], "el")
        self.assertEqual(language[1], "Greek")
        self.assertEqual(language, LANGUAGES.GREEK)

    def test_detect_italian(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Questo test verifica che il metodo funzioni correttamente"
        )
        self.assertEqual(language[0], "it")
        self.assertEqual(language[1], "Italian")
        self.assertEqual(language, LANGUAGES.ITALIAN)

    def test_detect_french(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Ce test vérifie que la méthode fonctionne correctement"
        )
        self.assertEqual(language[0], "fr")
        self.assertEqual(language[1], "French")
        self.assertEqual(language, LANGUAGES.FRENCH)

    def test_detect_ukrainian(self):
        language_detector = LanguageDetector()
        language = language_detector.detect(
            "Цей тест перевіряє, чи метод працює правильно"
        )
        self.assertEqual(language[0], "uk")
        self.assertEqual(language[1], "Ukrainian")
        self.assertEqual(language, LANGUAGES.UKRAINIAN)


if __name__ == "__main__":
    unittest.main()
